# Copyright: (c) 2018, Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import base64
import logging
import re
import typing
import warnings

import spnego
import spnego.channel_bindings
from cryptography import x509
from cryptography.exceptions import UnsupportedAlgorithm
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from requests.auth import AuthBase
from requests.packages.urllib3.response import HTTPResponse

from pypsrp._utils import get_hostname, to_bytes
from pypsrp.exceptions import AuthenticationError

log = logging.getLogger(__name__)


class NoCertificateRetrievedWarning(Warning):
    pass


class UnknownSignatureAlgorithmOID(Warning):
    pass


class HTTPNegotiateAuth(AuthBase):
    def __init__(
        self,
        username: typing.Optional[str] = None,
        password: typing.Optional[str] = None,
        auth_provider: str = "negotiate",
        send_cbt: bool = True,
        service: str = "WSMAN",
        delegate: bool = False,
        hostname_override: typing.Optional[str] = None,
        wrap_required: bool = False,
    ) -> None:
        """
        Creates a HTTP auth context that uses Microsoft's Negotiate protocol
        to complete the auth process. This currently only supports the NTLM
        and Kerberos providers in the Negotiate protocol.

        :param username: The username to authenticate with, if not specified
            this will be with the user currently logged in (Windows only) or
            the default Kerberos ticket in the cache
        :param password: The password for username, if not specified this will
            try to use implicit credentials available to the user
        :param auth_provider: The authentication provider to use
            'negotiate': Will try to use Kerberos if available and fallback to
                NTLM if that fails
            'ntlm': Will only use NTLM
            'kerberos': Will only use Kerberos and will fail if this is not
                available
        :param send_cbt: Try to bind the channel token (HTTPS only) to the auth
            process, default is True
        :param service: The service part of the SPN to authenticate with,
            defaults to HTTP
        :param delegate: Whether to get an auth token that allows the token to
            be delegated to other servers, this is only used with Kerberos and
            defaults to False
        :param hostname_override: Override the hostname used as part of the
            SPN, by default the hostname is based on the URL of the request
        :param wrap_required: Whether message encryption (wrapping) is
            required (controls what auth context is used)
        """
        self.username = username
        self.password = password
        self.auth_provider = auth_provider
        self.send_cbt = send_cbt
        self.service = service
        self.delegate = delegate
        self.hostname_override = hostname_override
        self.wrap_required = wrap_required
        self.contexts: typing.Dict[str, typing.Any] = {}

        self._regex = re.compile(r"(Kerberos|Negotiate|NTLM)\s*([^,]*),?", re.I)

    def __call__(self, request):
        request.headers["Connection"] = "Keep-Alive"
        request.register_hook("response", self.response_hook)

        return request

    def response_hook(self, response, **kwargs):
        if response.status_code == 401:
            matched_provider = self._check_auth_supported(response, ["Negotiate", "Kerberos", "NTLM"])
            kwargs["_pypsrp_auth_provider"] = matched_provider

            response = self.handle_401(response, **kwargs)

        return response

    def handle_401(self, response, **kwargs):
        response_auth_header = kwargs.pop("_pypsrp_auth_provider")
        response_auth_header_l = response_auth_header.lower()
        auth_provider = self.auth_provider

        if response_auth_header_l != self.auth_provider:
            if self.auth_provider == "negotiate":
                auth_provider = response_auth_header_l

            elif response_auth_header_l != "negotiate":
                raise ValueError(
                    "Server responded with the auth protocol '%s' which is incompatible with the "
                    "specified auth_provider '%s'" % (response_auth_header, auth_provider)
                )

        host = get_hostname(response.url)
        auth_hostname = self.hostname_override or host

        cbt = None
        if self.send_cbt:
            cbt_app_data = HTTPNegotiateAuth._get_cbt_data(response)
            if cbt_app_data:
                cbt = spnego.channel_bindings.GssChannelBindings(application_data=cbt_app_data)

        context_req = spnego.ContextReq.default
        if self.delegate:
            context_req |= spnego.ContextReq.delegate

        spnego_options = spnego.NegotiateOptions.wrapping_winrm if self.wrap_required else 0
        context = spnego.client(
            self.username,
            self.password,
            hostname=auth_hostname,
            service=self.service,
            channel_bindings=cbt,
            context_req=context_req,
            protocol=auth_provider,
            options=spnego_options,
        )
        self.contexts[host] = context

        out_token = context.step()
        while not context.complete or out_token is not None:
            # consume content and release the original connection to allow the
            # new request to reuse the same one.
            response.content
            response.raw.release_conn()

            # create a request with the Negotiate token present
            request = response.request.copy()
            log.debug("Sending http request with new auth token")
            self._set_auth_token(request, out_token, response_auth_header)

            # send the request with the auth token and get the response
            response = response.connection.send(request, **kwargs)

            # attempt to retrieve the auth token response
            in_token = self._get_auth_token(response, self._regex)

            # break if there was no token received from the host and return the
            # last response
            if in_token in [None, b""]:
                log.debug("Did not receive a http response with an auth response, stopping authentication process")
                break

            out_token = context.step(in_token)

        # This is used by the message encryption to decide the MIME protocol.
        setattr(context, "response_auth_header", response_auth_header_l)

        return response

    @staticmethod
    def _check_auth_supported(response, auth_providers):
        auth_supported = response.headers.get("www-authenticate", "")
        matched_providers = [p for p in auth_providers if p.upper() in auth_supported.upper()]
        if not matched_providers:
            raise AuthenticationError(
                "The server did not response with one of the following authentication methods "
                "%s - actual: '%s'" % (", ".join(auth_providers), auth_supported)
            )

        return matched_providers[0]

    @staticmethod
    def _set_auth_token(request, token, auth_provider):
        encoded_token = base64.b64encode(token)
        auth_header = to_bytes("%s " % auth_provider) + encoded_token
        request.headers["Authorization"] = auth_header

    @staticmethod
    def _get_auth_token(response, pattern):
        auth_header = response.headers.get("www-authenticate", "")
        token_match = pattern.search(auth_header)

        if not token_match:
            return None

        token = token_match.group(2)
        return base64.b64decode(token)

    @staticmethod
    def _get_cbt_data(response):
        """
        Tries to get the channel binding token as specified in RFC 5929 to pass
        along to the authentication provider. This is usually the SHA256
        hash of the certificate of the HTTPS endpoint appended onto the string
        'tls-server-end-point'.

        If the socket is not an SSL socker or the raw HTTP object is not a
        urllib3 HTTPResponse, then None will be returned and no channel binding
        data is passed onto the auth context

        :param response: The server's response which is used to sniff out the
            server's certificate
        :return: A byte string containing the CBT prefix and cert hash to pass
            onto the auth context
        """
        app_data = None
        raw_response = response.raw

        if isinstance(raw_response, HTTPResponse):
            try:
                socket = raw_response._fp.fp.raw._sock
            except AttributeError as err:
                warning = "Failed to get raw socket for CBT from urllib3 resp: %s" % str(err)
                warnings.warn(warning, NoCertificateRetrievedWarning)

            else:
                try:
                    cert = socket.getpeercert(True)
                except AttributeError:
                    pass
                else:
                    cert_hash = HTTPNegotiateAuth._get_certificate_hash(cert)
                    app_data = b"tls-server-end-point:" + cert_hash
        else:
            warning = (
                "Requests is running with a non urllib3 backend, cannot retrieve server cert for CBT. Raw "
                "type: %s" % type(response).__name__
            )
            warnings.warn(warning, NoCertificateRetrievedWarning)

        return app_data

    @staticmethod
    def _get_certificate_hash(certificate_der):
        """
        Get's the server's certificate hash for the tls-server-end-point
        channel binding.

        According to https://tools.ietf.org/html/rfc5929#section-4.1, this is
        calculated by
            Using the SHA256 is the signatureAlgorithm is MD5 or SHA1
            The signatureAlgorithm if the hash function is neither MD5 or SHA1

        :param certificate_der: The byte string of the server's certificate
        :return: The byte string containing the hash of the server's
            certificate
        """
        backend = default_backend()

        cert = x509.load_der_x509_certificate(certificate_der, backend)

        hash_algorithm = None
        try:
            hash_algorithm = cert.signature_hash_algorithm
        except UnsupportedAlgorithm as ex:
            warnings.warn(
                "Failed to get the signature algorithm from the certificate due to: %s" % str(ex),
                UnknownSignatureAlgorithmOID,
            )

        # If the cert signature algorithm is unknown, md5, or sha1 then use sha256 otherwise use the signature
        # algorithm of the cert itself.
        if not hash_algorithm or hash_algorithm.name in ["md5", "sha1"]:
            digest = hashes.Hash(hashes.SHA256(), backend)
        else:
            digest = hashes.Hash(hash_algorithm, backend)

        digest.update(certificate_der)
        certificate_hash = digest.finalize()

        return certificate_hash



import uuid
import xml.etree.ElementTree as ET
from queue import Empty, Queue

import pytest

from pypsrp._utils import to_string, to_unicode
from pypsrp.complex_objects import (
    ComplexObject,
    GenericComplexObject,
    ListMeta,
    ObjectMeta,
    StackMeta,
)
from pypsrp.exceptions import SerializationError
from pypsrp.serializer import Serializer, TaggedValue

from . import assert_xml_diff


class TestSerializer(object):
    @pytest.mark.parametrize(
        "input_val, expected",
        [
            ["0123456789abcdefghijklmnopqrstuvwxyz_", "0123456789abcdefghijklmnopqrstuvwxyz_"],
            ["actual_x000A_string\nnewline", "actual_x005F_x000A_string_x000A_newline"],
            ["treble clef %s" % b"\xd8\x34\xdd\x1e".decode("utf-16-be"), "treble clef _xD834__xDD1E_"],
            [None, None],
            ["upper_X000a_string\nnewline", "upper_x005F_X000a_string_x000A_newline"],
        ],
    )
    def test_serialize_string(self, input_val, expected):
        serializer = Serializer()

        actual_serial = serializer._serialize_string(input_val)
        assert actual_serial == expected
        actual_deserial = serializer._deserialize_string(actual_serial)
        assert actual_deserial == (input_val or "")

    @pytest.mark.parametrize(
        "data, expected",
        [
            ["a", "<S>a</S>"],
            ["a", "<S>a</S>"],
            [1, "<I32>1</I32>"],
            [True, "<B>true</B>"],
            [False, "<B>false</B>"],
            [10.0323, "<Sg>10.0323</Sg>"],
            [uuid.UUID(bytes=b"\x00" * 16), "<G>00000000-0000-0000-0000-000000000000</G>"],
            [TaggedValue("U32", 1), "<U32>1</U32>"],
        ],
    )
    def test_serialize_primitives(self, data, expected):
        serializer = Serializer()

        actual = serializer.serialize(data)
        actual_xml = to_unicode(ET.tostring(actual))
        expected_xml = to_unicode(expected)
        assert actual_xml == expected_xml

        deserial_actual = serializer.deserialize(actual)
        if isinstance(data, TaggedValue):
            data = data.value
        assert deserial_actual == data

    def test_serialize_byte_string_py3(self):
        serialzier = Serializer()
        expected = "<BA>YWJj</BA>"
        actual = serialzier.serialize(b"abc")
        actual_xml = to_unicode(ET.tostring(actual))
        expected_xml = to_unicode(expected)
        assert actual_xml == expected_xml

    def test_serialize_queue(self):
        serializer = Serializer()
        data = Queue()
        data.put("0")
        data.put("1")
        data.put("2")
        expected = (
            '<Obj RefId="0"><TN RefId="0"><T>System.Collections.Queue</T>'
            "<T>System.Object</T></TN>"
            "<QUE><S>0</S><S>1</S><S>2</S></QUE></Obj>"
        )
        actual = serializer.serialize(data)
        actual_xml = to_string(ET.tostring(actual))
        assert actual_xml == expected

        deserial_actual = serializer.deserialize(actual)
        assert isinstance(deserial_actual, Queue)
        assert deserial_actual.get() == "0"
        assert deserial_actual.get() == "1"
        assert deserial_actual.get() == "2"
        with pytest.raises(Empty):
            deserial_actual.get(block=False)

    def test_serialize_stack(self):
        # Python doesn't have a native type so serializing a list will reverse
        # the entries as a stack is last in first out. When deserializing we
        # don't reverse the entries as .pop(0) should return the last entry
        serializer = Serializer()
        data = []
        data.append("0")
        data.append("1")
        data.append("2")
        expected = (
            '<Obj RefId="0"><TN RefId="0"><T>System.Collections.Stack</T>'
            "<T>System.Object</T></TN>"
            "<STK><S>2</S><S>1</S><S>0</S></STK></Obj>"
        )

        actual = serializer.serialize(data, StackMeta())
        actual_xml = to_string(ET.tostring(actual))
        assert actual_xml == expected

        deserial_actual = serializer.deserialize(actual)
        assert deserial_actual == ["2", "1", "0"]

    def test_serialize_list(self):
        serializer = Serializer()
        data = []
        data.append("0")
        data.append("1")
        data.append("2")
        expected = (
            '<Obj RefId="0"><TN RefId="0"><T>System.Object[]</T>'
            "<T>System.Array</T><T>System.Object</T></TN>"
            "<LST><S>0</S><S>1</S><S>2</S></LST></Obj>"
        )

        actual = serializer.serialize(data)
        actual_xml = to_string(ET.tostring(actual))
        assert actual_xml == expected

        deserial_actual = serializer.deserialize(actual)
        assert deserial_actual == ["0", "1", "2"]

    def test_serialize_list_as_ie(self):
        serializer = Serializer()
        data = []
        data.append("0")
        data.append("1")
        data.append("2")
        expected = (
            '<Obj RefId="0"><TN RefId="0"><T>System.Object[]</T>'
            "<T>System.Array</T><T>System.Object</T></TN>"
            "<IE><S>0</S><S>1</S><S>2</S></IE></Obj>"
        )

        actual = serializer.serialize(data, ListMeta("IE"))
        actual_xml = to_string(ET.tostring(actual))
        assert actual_xml == expected

    def test_serialize_secure_string_no_cipher(self):
        serializer = Serializer()
        with pytest.raises(SerializationError) as err:
            serializer._serialize_secure_string("")
        assert str(err.value) == "Cannot generate secure string as cipher is not initialised"

    def test_deserialize_secure_string_no_cipher(self):
        serializer = Serializer()
        # should just return the input var
        actual = serializer._deserialize_secure_string("a")
        assert actual == "a"

    def test_serialize_dynamic_obj(self):
        serializer = Serializer()
        expected = '<Obj RefId="0"><MS><S N="key">value</S></MS></Obj>'
        obj = GenericComplexObject()
        obj.extended_properties["key"] = "value"
        actual = serializer.serialize(obj)
        actual_xml = to_string(ET.tostring(actual))
        assert actual_xml == expected

    def test_serialize_dynamic_complex(self):
        serializer = Serializer()
        expected = (
            '<Obj RefId="0"><TN RefId="0">'
            "<T>System.Management.Automation.PSCustomObject</T>"
            "<T>System.Object</T></TN><ToString>to string value</ToString>"
            '<I32>1</I32><S>2</S><MS><S N="extended_key">extended</S></MS>'
            '<Props><S N="adapted_key">adapted</S></Props></Obj>'
        )

        obj = GenericComplexObject()
        obj.types = ["System.Management.Automation.PSCustomObject", "System.Object"]
        obj.extended_properties["extended_key"] = "extended"
        obj.adapted_properties["adapted_key"] = "adapted"
        obj.property_sets = [1, "2"]
        obj.to_string = "to string value"

        actual = serializer.serialize(obj)
        actual_xml = to_string(ET.tostring(actual))
        assert actual_xml == expected

    def test_deserialize_obj_no_types(self):
        serializer = Serializer()
        xml = '<Obj RefId="0"><MS><S N="key">value</S></MS></Obj>'
        actual = serializer.deserialize(xml)
        assert isinstance(actual, GenericComplexObject)
        assert actual.adapted_properties == {}
        assert actual.extended_properties == {"key": "value"}
        assert actual.property_sets == []
        assert actual.to_string is None
        assert actual.types == []

    def test_deserialize_obj_missing_prop(self):
        class SerialObject(ComplexObject):
            def __init__(self, **kwargs):
                super(SerialObject, self).__init__()
                self._types = ["System.Test", "System.Object"]
                self._extended_properties = (("man_prop", ObjectMeta("S", optional=False)),)
                self.man_prop = kwargs.get("man_prop")

        serializer = Serializer()
        xml = '<Obj RefId="0"><TN RefId="0"><T>System.Test</T>' "<T>System.Object</T></TN><MS /></Obj>"
        with pytest.raises(SerializationError) as err:
            serializer.deserialize(xml, ObjectMeta("Obj", object=SerialObject))
        assert str(err.value) == "Mandatory return value for 'Unknown' was not found on object Unknown"

    def test_deserialize_obj_missing_prop_names(self):
        class SerialObject(ComplexObject):
            def __init__(self, **kwargs):
                super(SerialObject, self).__init__()
                self._types = ["System.Test", "System.Object"]
                self._extended_properties = (("man_prop", ObjectMeta("S", name="key", optional=False)),)
                self.man_prop = kwargs.get("man_prop")

        serializer = Serializer()
        xml = (
            '<Obj RefId="0"><TN RefId="0"><T>System.Test</T>'
            "<T>System.Object</T></TN><ToString>obj</ToString><MS /></Obj>"
        )
        with pytest.raises(SerializationError) as err:
            serializer.deserialize(xml, ObjectMeta("Obj", object=SerialObject))
        assert str(err.value) == "Mandatory return value for 'key' was not found on object obj"

    def test_deserialize_dynamic_obj_type_ref(self):
        serializer = Serializer()
        xml1 = (
            '<Obj RefId="0"><TN RefId="0">'
            "<T>System.Management.Automation.PSCustomObject</T>"
            "<T>System.Object</T></TN><ToString>to string value</ToString>"
            '<I32>1</I32><S>2</S><MS><S N="extended_key">extended</S></MS>'
            '<Props><S N="adapted_key">adapted</S></Props></Obj>'
        )
        xml2 = (
            '<Obj RefId="2"><TNRef RefId="0" />'
            "<ToString>to string value 2</ToString>"
            '<I32>1</I32><S>2</S><MS><S N="extended_key">extended</S></MS>'
            '<Props><S N="adapted_key">adapted</S></Props></Obj>'
        )
        serializer.deserialize(xml1)
        actual = serializer.deserialize(xml2, clear=False)
        assert str(actual) == "to string value 2"
        assert actual.types == ["System.Management.Automation.PSCustomObject", "System.Object"]

    def test_deserialize_empty_string(self):
        serializer = Serializer()
        actual = serializer.deserialize("")
        assert actual == ""

    def test_deserialize_unknown_tag(self):
        serializer = Serializer()
        xml = """<Obj N="Value" RefId="14">
    <MS>
        <S N="T">System.Management.Automation.Host.Size</S>
        <Obj N="V" RefId="15">
            <MS>
                <I32 N="height">50</I32>
                <I32 N="width">60</I32>
            </MS>
        </Obj>
    </MS>
</Obj>"""
        actual = serializer.deserialize(xml, ObjectMeta("fake", object=GenericComplexObject))
        assert actual == xml

    def test_serialize_circualr_reference(self):
        serializer = Serializer()
        obj = GenericComplexObject()
        obj.types = [
            "Microsoft.Exchange.Data.Directory.ADObjectId",
            "Microsoft.Exchange.Data.ObjectId",
            "System.Object",
        ]
        obj.to_string = "com"
        obj.adapted_properties = {
            "OrgHierarchyToIgnore": None,
            "IsDeleted": False,
            "Parent": None,
            "Depth": 0,
            "DistinguishedName": "DC=com",
            "IsRelativeDn": False,
            "DomainId": obj,
            "Name": "com",
            "SecurityIdentifierString": None,
        }
        obj.property_sets.append("abc")
        obj.property_sets.append(obj)

        expected = (
            '<Obj RefId="0"><TN RefId="0"><T>Microsoft.Exchange.Data.'
            "Directory.ADObjectId</T><T>Microsoft.Exchange.Data.ObjectId</T>"
            "<T>System.Object</T></TN><ToString>com</ToString><S>abc</S>"
            '<Ref RefId="0" /><Props><I32 N="Depth">0</I32>'
            '<S N="DistinguishedName">DC=com</S>'
            '<Ref N="DomainId" RefId="0" /><B N="IsDeleted">false</B>'
            '<B N="IsRelativeDn">false</B><S N="Name">com</S>'
            '<Nil N="OrgHierarchyToIgnore" /><Nil N="Parent" />'
            '<Nil N="SecurityIdentifierString" /></Props></Obj>'
        )
        actual = serializer.serialize(obj)
        actual_xml = to_string(ET.tostring(actual))
        assert_xml_diff(actual_xml, expected)

    def test_deserialize_circular_reference(self):
        serializer = Serializer()
        xml = """<Obj N="DomainId" RefId="44">
    <TN RefId="6">
        <T>Microsoft.Exchange.Data.Directory.ADObjectId</T>
        <T>Microsoft.Exchange.Data.ObjectId</T>
        <T>System.Object</T>
    </TN>
    <ToString>com</ToString>
    <Props>
        <Nil N="OrgHierarchyToIgnore"/>
        <B N="IsDeleted">false</B>
        <Nil N="Parent"/>
        <I32 N="Depth">0</I32>
        <S N="DistinguishedName">DC=com</S>
        <B N="IsRelativeDn">false</B>
        <Ref N="DomainId" RefId="44"/>
        <S N="Name">com</S>
        <Nil N="SecurityIdentifierString"/>
    </Props>
</Obj>"""
        actual = serializer.deserialize(xml)
        assert str(actual) == "com"
        assert str(actual.adapted_properties["DomainId"]) == "com"
        assert actual.adapted_properties["DomainId"].types == [
            "Microsoft.Exchange.Data.Directory.ADObjectId",
            "Microsoft.Exchange.Data.ObjectId",
            "System.Object",
        ]



import logging
import re
import struct
import typing

from pypsrp._utils import to_bytes
from pypsrp.exceptions import WinRMError

log = logging.getLogger(__name__)


class WinRMEncryption(object):

    SIXTEEN_KB = 16384
    MIME_BOUNDARY = "--Encrypted Boundary"
    CREDSSP = "application/HTTP-CredSSP-session-encrypted"
    KERBEROS = "application/HTTP-Kerberos-session-encrypted"
    SPNEGO = "application/HTTP-SPNEGO-session-encrypted"

    def __init__(self, context: typing.Any, protocol: str) -> None:
        log.debug("Initialising WinRMEncryption helper for protocol %s" % protocol)
        self.context = context
        self.protocol = protocol

    def wrap_message(self, message: bytes) -> typing.Tuple[str, bytes]:
        log.debug("Wrapping message")
        if self.protocol == self.CREDSSP and len(message) > self.SIXTEEN_KB:
            content_type = "multipart/x-multi-encrypted"
            encrypted_msg = b""
            chunks = [message[i : i + self.SIXTEEN_KB] for i in range(0, len(message), self.SIXTEEN_KB)]
            for chunk in chunks:
                encrypted_chunk = self._wrap_message(chunk)
                encrypted_msg += encrypted_chunk
        else:
            content_type = "multipart/encrypted"
            encrypted_msg = self._wrap_message(message)

        encrypted_msg += to_bytes("%s--\r\n" % self.MIME_BOUNDARY)

        log.debug("Created wrapped message of content type %s" % content_type)
        return content_type, encrypted_msg

    def unwrap_message(self, message: bytes, boundary: str) -> bytes:
        log.debug("Unwrapped message")

        # Talking to Exchange endpoints gives a non-compliant boundary that has a space between the -- {boundary}, not
        # ideal but we just need to handle it.
        parts = re.compile(to_bytes(r"--\s*%s\r\n" % re.escape(boundary))).split(message)
        parts = list(filter(None, parts))

        message = b""
        for i in range(0, len(parts), 2):
            header = parts[i].strip()
            payload = parts[i + 1]

            expected_length = int(header.split(b"Length=")[1])

            # remove the end MIME block if it exists
            payload = re.sub(to_bytes(r"--\s*%s--\r\n$") % to_bytes(boundary), b"", payload)

            wrapped_data = payload.replace(b"\tContent-Type: application/octet-stream\r\n", b"")
            header_length = struct.unpack("<i", wrapped_data[:4])[0]
            header = wrapped_data[4 : 4 + header_length]
            enc_wrapped_data = wrapped_data[4 + header_length :]

            unwrapped_data = self.context.unwrap_winrm(header, enc_wrapped_data)

            actual_length = len(unwrapped_data)

            log.debug("Actual unwrapped length: %d, expected unwrapped length: %d" % (actual_length, expected_length))
            if actual_length != expected_length:
                raise WinRMError(
                    "The encrypted length from the server does "
                    "not match the expected length, decryption "
                    "failed, actual: %d != expected: %d" % (actual_length, expected_length)
                )
            message += unwrapped_data

        return message

    def _wrap_message(self, message: bytes) -> bytes:
        header, wrapped_data, padding_length = self.context.wrap_winrm(message)
        wrapped_data = struct.pack("<i", len(header)) + header + wrapped_data

        msg_length = str(len(message) + padding_length)

        payload = "\r\n".join(
            [
                self.MIME_BOUNDARY,
                "\tContent-Type: %s" % self.protocol,
                "\tOriginalContent: type=application/soap+xml;charset=UTF-8;Length=%s" % msg_length,
                self.MIME_BOUNDARY,
                "\tContent-Type: application/octet-stream",
                "",
            ]
        )
        return to_bytes(payload) + wrapped_data



import uuid

from pypsrp.complex_objects import HostMethodIdentifier, ObjectMeta
from pypsrp.messages import (
    ErrorRecordMessage,
    Message,
    MessageType,
    PublicKeyRequest,
    RunspacePoolHostCall,
    RunspacePoolHostResponse,
    UserEvent,
    WarningRecord,
)
from pypsrp.serializer import Serializer


class TestPublicKeyRequest(object):
    def test_create_public_key_request(self):
        pub_key_req = PublicKeyRequest()
        empty_uuid = "00000000-0000-0000-0000-000000000000"
        serializer = Serializer()
        expected = b"<S />"

        msg = Message(0x2, empty_uuid, empty_uuid, pub_key_req, serializer)
        actual = msg.pack()
        assert (
            actual == b"\x02\x00\x00\x00"
            b"\x07\x00\x01\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00" + expected
        )

    def test_parse_public_key_request(self):
        data = (
            b"\x02\x00\x00\x00"
            b"\x07\x00\x01\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"\x00\x00\x00\x00\x00\x00\x00\x00"
            b"<S />"
        )
        actual = Message.unpack(data, Serializer())
        assert actual.message_type == MessageType.PUBLIC_KEY_REQUEST
        assert isinstance(actual.data, PublicKeyRequest)


class TestUserEvent(object):
    def test_parse_msg(self):
        xml = """<Obj RefId="0">
            <MS>
                <I32 N="PSEventArgs.EventIdentifier">1</I32>
                <S N="PSEventArgs.SourceIdentifier">ae6245f2-c179-4a9a-a039-47b60fc44500</S>
                <DT N="PSEventArgs.TimeGenerated">2009-06-17T10:57:23.1578277-07:00</DT>
                <Obj N="PSEventArgs.Sender" RefId="1">
                    <TN RefId="0">
                        <T>System.Timers.Timer</T>
                        <T>System.ComponentModel.Component</T>
                        <T>System.MarshalByRefObject</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>System.Timers.Timer</ToString>
                    <Props>
                        <B N="AutoReset">true</B>
                        <B N="Enabled">true</B>
                        <Db N="Interval">5000</Db>
                        <Nil N="Site"/>
                        <Nil N="SynchronizingObject"/>
                        <Nil N="Container"/>
                    </Props>
                </Obj>
                <Obj N="PSEventArgs.SourceArgs" RefId="2">
                    <TN RefId="1">
                        <T>System.Object[]</T>
                        <T>System.Array</T>
                        <T>System.Object</T>
                    </TN>
                    <LST>
                        <Ref RefId="1"/>
                        <Obj RefId="3">
                            <TN RefId="2">
                                <T>System.Timers.ElapsedEventArgs</T>
                                <T>System.EventArgs</T>
                                <T>System.Object</T>
                            </TN>
                            <ToString>System.Timers.ElapsedEventArgs</ToString>
                            <Props>
                                <DT N="SignalTime">2009-06-17T10:57:23.1568275-07:00</DT>
                            </Props>
                        </Obj>
                    </LST>
                </Obj>
                <Nil N="PSEventArgs.MessageData"/>
                <Nil N="PSEventArgs.ComputerName"/>
                <G N="PSEventArgs.RunspaceId">fb9c87e8-1190-40a7-a681-6fc9b9f84a17</G>
            </MS>
        </Obj>"""
        serializer = Serializer()
        meta = ObjectMeta("Obj", object=UserEvent)
        actual = serializer.deserialize(xml, meta)

        assert str(actual.args[0]) == "System.Timers.Timer"
        assert actual.args[0].adapted_properties["Interval"] == 5000.0
        assert str(actual.args[1]) == "System.Timers.ElapsedEventArgs"
        assert actual.args[1].adapted_properties["SignalTime"] == "2009-06-17T10:57:23.1568275-07:00"
        assert actual.computer is None
        assert actual.data is None
        assert actual.event_id == 1
        assert actual.runspace_id == uuid.UUID("fb9c87e8-1190-40a7-a681-6fc9b9f84a17")
        assert str(actual.sender) == "System.Timers.Timer"
        assert actual.sender.adapted_properties["Interval"] == 5000.0
        assert actual.source_id == "ae6245f2-c179-4a9a-a039-47b60fc44500"
        assert actual.time == "2009-06-17T10:57:23.1578277-07:00"


class TestRunspacePoolHostCall(object):
    def test_parse_message(self):
        xml = """<Obj RefId="0">
            <MS>
                <I64 N="ci">1</I64>
                <Obj N="mi" RefId="1">
                    <TN RefId="0">
                        <T>System.Management.Automation.Remoting.RemoteHostMethodId</T>
                        <T>System.Enum</T>
                        <T>System.ValueType</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>ReadLine</ToString>
                    <I32>11</I32>
                </Obj>
                <Obj N="mp" RefId="2">
                    <TN RefId="1">
                        <T>System.Collections.ArrayList</T>
                        <T>System.Object</T>
                    </TN>
                    <LST/>
                </Obj>
            </MS>
        </Obj>"""
        serializer = Serializer()
        meta = ObjectMeta("Obj", object=RunspacePoolHostCall)
        actual = serializer.deserialize(xml, meta)
        assert actual.ci == 1
        assert str(actual.mi) == "ReadLine"
        assert isinstance(actual.mi, HostMethodIdentifier)
        assert actual.mi.value == 11
        assert actual.mp == []


class TestRunspacePoolHostResponse(object):
    def test_parse_message(self):
        xml = """<Obj RefId="11">
            <MS>
                <S N="mr">Line read from the host</S>
                <I64 N="ci">1</I64>
                <Obj N="mi" RefId="12">
                    <TN RefId="4">
                        <T>System.Management.Automation.Remoting.RemoteHostMethodId</T>
                        <T>System.Enum</T>
                        <T>System.ValueType</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>ReadLine</ToString>
                    <I32>11</I32>
                </Obj>
            </MS>
        </Obj>"""
        serializer = Serializer()
        meta = ObjectMeta("Obj", object=RunspacePoolHostResponse)
        actual = serializer.deserialize(xml, meta)
        assert actual.ci == 1
        assert actual.me is None
        assert str(actual.mi) == "ReadLine"
        assert isinstance(actual.mi, HostMethodIdentifier)
        assert actual.mi.value == 11
        assert actual.mr == "Line read from the host"


class TestErrorRecord(object):
    def test_parse_error(self):
        xml = """<Obj RefId="0">
            <TN RefId="0">
                <T>System.Management.Automation.ErrorRecord</T>
                <T>System.Object</T>
            </TN>
            <ToString>error stream</ToString>
            <MS>
                <B N="writeErrorStream">true</B>
                <Obj N="Exception" RefId="1">
                    <TN RefId="1">
                        <T>Microsoft.PowerShell.Commands.WriteErrorException</T>
                        <T>System.SystemException</T>
                        <T>System.Exception</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>Microsoft.PowerShell.Commands.WriteErrorException: error stream</ToString>
                    <Props>
                        <S N="Message">error stream</S>
                        <Obj N="Data" RefId="2">
                            <TN RefId="2">
                                <T>System.Collections.ListDictionaryInternal</T>
                                <T>System.Object</T>
                            </TN>
                            <DCT/>
                        </Obj>
                        <Nil N="InnerException"/>
                        <Nil N="TargetSite"/>
                        <Nil N="StackTrace"/>
                        <Nil N="HelpLink"/>
                        <Nil N="Source"/>
                        <I32 N="HResult">-2146233087</I32>
                    </Props>
                </Obj>
                <Nil N="TargetObject"/>
                <S N="FullyQualifiedErrorId">Microsoft.PowerShell.Commands.WriteErrorException</S>
                <Obj N="InvocationInfo" RefId="3">
                    <TN RefId="3">
                        <T>System.Management.Automation.InvocationInfo</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>System.Management.Automation.InvocationInfo</ToString>
                    <Props>
                        <S N="MyCommand">Write-Error 'error stream'_x000A_</S>
                        <Obj N="BoundParameters" RefId="4">
                            <TN RefId="4">
                                <T>System.Collections.Generic.Dictionary`2[[System.String, mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089],[System.Object, mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089]]</T>
                                <T>System.Object</T>
                            </TN>
                            <DCT/>
                        </Obj>
                        <Obj N="UnboundArguments" RefId="5">
                            <TN RefId="5">
                                <T>System.Collections.Generic.List`1[[System.Object, mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089]]</T>
                                <T>System.Object</T>
                            </TN>
                            <LST/>
                        </Obj>
                        <I32 N="ScriptLineNumber">0</I32>
                        <I32 N="OffsetInLine">0</I32>
                        <I64 N="HistoryId">1</I64>
                        <S N="ScriptName"/>
                        <S N="Line"/>
                        <S N="PositionMessage"/>
                        <S N="PSScriptRoot"/>
                        <Nil N="PSCommandPath"/>
                        <S N="InvocationName"/>
                        <I32 N="PipelineLength">0</I32>
                        <I32 N="PipelinePosition">0</I32>
                        <B N="ExpectingInput">false</B>
                        <S N="CommandOrigin">Internal</S>
                        <Nil N="DisplayScriptPosition"/>
                    </Props>
                </Obj>
                <I32 N="ErrorCategory_Category">0</I32>
                <S N="ErrorCategory_Activity">Write-Error</S>
                <S N="ErrorCategory_Reason">WriteErrorException</S>
                <S N="ErrorCategory_TargetName"/>
                <S N="ErrorCategory_TargetType"/>
                <S N="ErrorCategory_Message">NotSpecified: (:) [Write-Error], WriteErrorException</S>
                <B N="SerializeExtendedInfo">true</B>
                <Ref N="InvocationInfo_BoundParameters" RefId="4"/>
                <Obj N="InvocationInfo_CommandOrigin" RefId="6">
                    <TN RefId="6">
                        <T>System.Management.Automation.CommandOrigin</T>
                        <T>System.Enum</T>
                        <T>System.ValueType</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>Internal</ToString>
                    <I32>1</I32>
                </Obj>
                <B N="InvocationInfo_ExpectingInput">false</B>
                <S N="InvocationInfo_InvocationName"/>
                <S N="InvocationInfo_Line"/>
                <I32 N="InvocationInfo_OffsetInLine">0</I32>
                <I64 N="InvocationInfo_HistoryId">1</I64>
                <Obj N="InvocationInfo_PipelineIterationInfo" RefId="7">
                    <TN RefId="7">
                        <T>System.Int32[]</T>
                        <T>System.Array</T>
                        <T>System.Object</T>
                    </TN>
                    <LST/>
                </Obj>
                <I32 N="InvocationInfo_PipelineLength">0</I32>
                <I32 N="InvocationInfo_PipelinePosition">0</I32>
                <S N="InvocationInfo_PSScriptRoot"/>
                <Nil N="InvocationInfo_PSCommandPath"/>
                <S N="InvocationInfo_PositionMessage"/>
                <I32 N="InvocationInfo_ScriptLineNumber">0</I32>
                <S N="InvocationInfo_ScriptName"/>
                <Ref N="InvocationInfo_UnboundArguments" RefId="5"/>
                <B N="SerializeExtent">false</B>
                <Obj N="CommandInfo_CommandType" RefId="8">
                    <TN RefId="8">
                        <T>System.Management.Automation.CommandTypes</T>
                        <T>System.Enum</T>
                        <T>System.ValueType</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>Script</ToString>
                    <I32>64</I32>
                </Obj>
                <S N="CommandInfo_Definition">Write-Error 'error stream'_x000A_</S>
                <S N="CommandInfo_Name"/>
                <Obj N="CommandInfo_Visibility" RefId="9">
                    <TN RefId="9">
                        <T>System.Management.Automation.SessionStateEntryVisibility</T>
                        <T>System.Enum</T>
                        <T>System.ValueType</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>Public</ToString>
                    <I32>0</I32>
                </Obj>
                <Obj N="PipelineIterationInfo" RefId="10">
                    <TN RefId="10">
                        <T>System.Collections.ObjectModel.ReadOnlyCollection`1[[System.Int32, mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089]]</T>
                        <T>System.Object</T>
                    </TN>
                    <LST>
                        <I32>0</I32>
                        <I32>0</I32>
                    </LST>
                </Obj>
                <S N="ErrorDetails_ScriptStackTrace">at &lt;ScriptBlock&gt;, &lt;No file&gt;: line 5</S>
                <Nil N="PSMessageDetails"/>
            </MS>
        </Obj>"""
        serializer = Serializer()
        meta = ObjectMeta("Obj", object=ErrorRecordMessage)
        actual = serializer.deserialize(xml, meta)

        assert str(actual) == "error stream"
        assert actual.exception.adapted_properties["Message"] == "error stream"
        assert actual.exception.adapted_properties["HResult"] == -2146233087
        assert actual.target_object is None
        assert actual.fq_error == "Microsoft.PowerShell.Commands.WriteErrorException"
        assert actual.invocation
        invoc_props = actual.invocation_info.adapted_properties
        assert invoc_props["MyCommand"] == "Write-Error 'error stream'\n"
        assert invoc_props["BoundParameters"] == {}
        assert invoc_props["UnboundArguments"] == []
        assert invoc_props["ScriptLineNumber"] == 0
        assert invoc_props["OffsetInLine"] == 0
        assert invoc_props["HistoryId"] == 1
        assert invoc_props["CommandOrigin"] == "Internal"
        assert actual.fq_error == "Microsoft.PowerShell.Commands.WriteErrorException"
        assert actual.category == 0
        assert actual.reason == "WriteErrorException"
        assert actual.target_name == ""
        assert actual.target_type == ""
        assert actual.message == "NotSpecified: (:) [Write-Error], WriteErrorException"
        assert actual.details_message is None
        assert actual.action is None
        assert actual.script_stacktrace == "at <ScriptBlock>, <No file>: line 5"
        assert actual.extended_info_present
        assert actual.invocation_name == ""
        assert actual.invocation_bound_parameters == {}
        assert actual.invocation_unbound_arguments == []
        assert str(actual.invocation_command_origin) == "Internal"
        assert actual.invocation_expecting_input is False
        assert actual.invocation_line == ""
        assert actual.invocation_offset_in_line == 0
        assert actual.invocation_position_message == ""
        assert actual.invocation_script_name == ""
        assert actual.invocation_script_line_number == 0
        assert actual.invocation_history_id == 1
        assert actual.invocation_pipeline_length == 0
        assert actual.invocation_pipeline_position == 0
        assert actual.invocation_pipeline_iteration_info == []
        assert str(actual.command_type) == "Script"
        assert actual.command_definition == "Write-Error 'error stream'\n"
        assert actual.command_name == ""
        assert str(actual.command_visibility) == "Public"
        assert actual.pipeline_iteration_info == [0, 0]


class TestWarningRecord(object):
    def test_parse_warning(self):
        xml = """<Obj RefId="0">
            <TN RefId="0">
                <T>System.Management.Automation.WarningRecord</T>
                <T>System.Management.Automation.InformationalRecord</T>
                <T>System.Object</T>
            </TN>
            <ToString>warning stream</ToString>
            <MS>
                <S N="InformationalRecord_Message">warning stream</S>
                <B N="InformationalRecord_SerializeInvocationInfo">true</B>
                <Obj N="InvocationInfo_BoundParameters" RefId="1">
                    <TN RefId="1">
                        <T>System.Management.Automation.PSBoundParametersDictionary</T>
                        <T>System.Collections.Generic.Dictionary`2[[System.String, mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089],[System.Object, mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089]]</T>
                        <T>System.Object</T>
                    </TN>
                    <DCT/>
                </Obj>
                <Obj N="InvocationInfo_CommandOrigin" RefId="2">
                    <TN RefId="2">
                        <T>System.Management.Automation.CommandOrigin</T>
                        <T>System.Enum</T>
                        <T>System.ValueType</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>Runspace</ToString>
                    <I32>0</I32>
                </Obj>
                <B N="InvocationInfo_ExpectingInput">false</B>
                <S N="InvocationInfo_InvocationName"/>
                <S N="InvocationInfo_Line"/>
                <I32 N="InvocationInfo_OffsetInLine">0</I32>
                <I64 N="InvocationInfo_HistoryId">1</I64>
                <Obj N="InvocationInfo_PipelineIterationInfo" RefId="3">
                    <TN RefId="3">
                        <T>System.Int32[]</T>
                        <T>System.Array</T>
                        <T>System.Object</T>
                    </TN>
                    <LST>
                        <I32>0</I32>
                        <I32>0</I32>
                    </LST>
                </Obj>
                <I32 N="InvocationInfo_PipelineLength">1</I32>
                <I32 N="InvocationInfo_PipelinePosition">1</I32>
                <S N="InvocationInfo_PSScriptRoot"/>
                <Nil N="InvocationInfo_PSCommandPath"/>
                <S N="InvocationInfo_PositionMessage"/>
                <I32 N="InvocationInfo_ScriptLineNumber">0</I32>
                <S N="InvocationInfo_ScriptName"/>
                <Obj N="InvocationInfo_UnboundArguments" RefId="4">
                    <TN RefId="4">
                        <T>System.Collections.Generic.List`1[[System.Object, mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089]]</T>
                        <T>System.Object</T>
                    </TN>
                    <LST/>
                </Obj>
                <B N="SerializeExtent">false</B>
                <Obj N="CommandInfo_CommandType" RefId="5">
                    <TN RefId="5">
                        <T>System.Management.Automation.CommandTypes</T>
                        <T>System.Enum</T>
                        <T>System.ValueType</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>Script</ToString>
                    <I32>64</I32>
                </Obj>
                <S N="CommandInfo_Definition">Write-Warning 'warning stream'_x000A_</S>
                <S N="CommandInfo_Name"/>
                <Obj N="CommandInfo_Visibility" RefId="6">
                    <TN RefId="6">
                        <T>System.Management.Automation.SessionStateEntryVisibility</T>
                        <T>System.Enum</T>
                        <T>System.ValueType</T>
                        <T>System.Object</T>
                    </TN>
                    <ToString>Public</ToString>
                    <I32>0</I32>
                </Obj>
                <Obj N="InformationalRecord_PipelineIterationInfo" RefId="7">
                    <TN RefId="7">
                        <T>System.Collections.ObjectModel.ReadOnlyCollection`1[[System.Int32, mscorlib, Version=4.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089]]</T>
                        <T>System.Object</T>
                    </TN>
                    <LST>
                        <I32>0</I32>
                        <I32>0</I32>
                    </LST>
                </Obj>
            </MS>
        </Obj>"""
        serializer = Serializer()
        meta = ObjectMeta("Obj", object=WarningRecord)
        actual = serializer.deserialize(xml, meta)

        assert str(actual) == "warning stream"
        assert actual.invocation
        assert actual.message == "warning stream"
        assert actual.pipeline_iteration_info == [0, 0]
        assert actual.invocation_name == ""
        assert actual.invocation_bound_parameters == {}
        assert actual.invocation_unbound_arguments == []
        assert str(actual.invocation_command_origin) == "Runspace"
        assert actual.invocation_expecting_input is False
        assert actual.invocation_line == ""
        assert actual.invocation_offset_in_line == 0
        assert actual.invocation_position_message == ""
        assert actual.invocation_script_name == ""
        assert actual.invocation_script_line_number == 0
        assert actual.invocation_history_id == 1
        assert actual.invocation_pipeline_length == 1
        assert actual.invocation_pipeline_position == 1
        assert actual.invocation_pipeline_iteration_info == [0, 0]
        assert actual.command_definition == "Write-Warning 'warning stream'\n"
        assert actual.command_name == ""
        assert str(actual.command_type) == "Script"
        assert str(actual.command_visibility) == "Public"
